#
# Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
#
from pydantic import parse_raw_as
import requests

from aistore.client.errors import AISError, ErrBckNotFound, ErrRemoteBckNotFound
from aistore.client.types import HttpError


def _raise_error(text: str):
    err = parse_raw_as(HttpError, text)
    if 400 <= err.status < 500:
        err = parse_raw_as(HttpError, text)
        if "does not exist" in err.message:
            if "cloud bucket" in err.message or "remote bucket" in err.message:
                raise ErrRemoteBckNotFound(err.status, err.message)
            if "bucket" in err.message:
                raise ErrBckNotFound(err.status, err.message)
    raise AISError(err.status, err.message)


# pylint: disable=unused-variable
def handle_errors(resp: requests.Response):
    """
    Error handling for requests made to the AIS Client

    Args:
        resp: requests.Response = Response received from the request
    """
    error_text = resp.text
    if isinstance(resp.text, bytes):
        try:
            error_text = error_text.decode("utf-8")
        except UnicodeDecodeError:
            error_text = error_text.decode("iso-8859-1")
    if error_text != "":
        _raise_error(error_text)
    resp.raise_for_status()


def probing_frequency(dur: int) -> float:
    freq = min(dur / 8.0, 1.0)
    freq = max(dur / 64.0, freq)
    return max(freq, 0.1)



#
# Copyright (c) 2018-2022, NVIDIA CORPORATION. All rights reserved.
#


class AISError(Exception):
    def __init__(self, status_code: int, message: str):
        self.status_code = status_code
        self.message = message
        super().__init__(f"STATUS:{status_code}, MESSAGE:{message}")


# pylint: disable=unused-variable
class InvalidBckProvider(Exception):
    def __init__(self, provider):
        super().__init__(f"Invalid bucket provider {provider}")


# pylint: disable=unused-variable
class ErrRemoteBckNotFound(AISError):
    def __init__(self, status_code, message):
        super().__init__(status_code=status_code, message=message)


# pylint: disable=unused-variable
class ErrBckNotFound(AISError):
    def __init__(self, status_code, message):
        super().__init__(status_code=status_code, message=message)


# pylint: disable=unused-variable
class Timeout(Exception):
    def __init__(self, action):
        super().__init__(f"{action} timed out")



#
# Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
#

from __future__ import annotations  # pylint: disable=unused-variable
from typing import List
import time

from aistore.client.const import HTTP_METHOD_GET, HTTP_METHOD_PUT, QParamWhat
from aistore.client.errors import Timeout
from aistore.client.types import Bck, XactStatus
from aistore.client.utils import probing_frequency


# pylint: disable=unused-variable
class Xaction:
    """
    A class containing xaction-related functions.

    Args:
        None
    """

    def __init__(self, client):
        self._client = client

    @property
    def client(self):
        """The client bound to this xaction object."""
        return self._client

    def xact_status(
        self,
        xact_id: str = "",
        xact_kind: str = "",
        daemon_id: str = "",
        only_running: bool = False,
    ) -> XactStatus:
        """
        Return status of an eXtended Action (xaction)

        Args:
            xact_id (str, optional): UUID of the xaction. Empty - all xactions.
            xact_kind (str, optional): Kind of the xaction. Empty - all kinds.
            daemon_id (str, optional): Return xactions only running on the daemon_id.
            only_running (bool, optional): True - return only currently running xactions, False - include in the list also finished and aborted ones.

        Returns:
            The xaction description.

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
        """
        value = {
            "id": xact_id,
            "kind": xact_kind,
            "show_active": only_running,
            "node": daemon_id,
        }
        params = {QParamWhat: "status"}

        return self.client.request_deserialize(
            HTTP_METHOD_GET,
            path="cluster",
            res_model=XactStatus,
            json=value,
            params=params,
        )

    def wait_for_xaction_finished(
        self,
        xact_id: str = "",
        xact_kind: str = "",
        daemon_id: str = "",
        timeout: int = 300,
    ):
        """
        Wait for an eXtended Action (xaction) to finish

        Args:
            xact_id (str, optional): UUID of the xaction. Empty - all xactions.
            xact_kind (str, optional): Kind of the xaction. Empty - all kinds.
            daemon_id (str, optional): Return xactions only running on the daemon_id.
            timeout (int, optional): The maximum time to wait for the xaction, in seconds. Default timeout is 5 minutes.

        Returns:
            None

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            errors.Timeout: Timeout while waiting for the xaction to finish
        """
        passed = 0
        sleep_time = probing_frequency(timeout)
        while True:
            if passed > timeout:
                raise Timeout("wait for xaction to finish")
            status = self.xact_status(
                xact_id=xact_id, xact_kind=xact_kind, daemon_id=daemon_id
            )
            if status.end_time != 0:
                break
            time.sleep(sleep_time)
            passed += sleep_time
            print(status)

    def xact_start(
        self,
        xact_kind: str = "",
        daemon_id: str = "",
        force: bool = False,
        buckets: List[Bck] = None,
    ) -> str:
        """
        Start an eXtended Action (xaction) and return its UUID.

        Args:
            xact_kind (str, optional): Kind of the xaction (for supported kinds, see api/apc/const.go). Empty - all kinds.
            daemon_id (str, optional): Return xactions only running on the daemon_id.
            force (bool, optional): Override existing restrictions for a bucket (e.g., run LRU eviction even if the bucket has LRU disabled).
            buckets (List[Bck], optional): List of one or more buckets; applicable only for xactions that have bucket scope (for details and full enumeration, see xact/table.go).

        Returns:
            The running xaction UUID.

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
        """
        value = {"kind": xact_kind, "node": daemon_id, "buckets": buckets}
        if force:
            value["ext"] = {"force": True}
        action = {"action": "start", "value": value}

        resp = self.client.request(
            HTTP_METHOD_PUT,
            path="cluster",
            json=action,
        )
        return resp.text



# returns the md5 sum of the original data as the response
# pylint: disable=unused-variable
MD5 = """
apiVersion: v1
kind: Pod
metadata:
  name: transformer-md5
  annotations:
    # Values it can take ["hpull://","hrev://","hpush://"]
    communication_type: "{communication_type}://"
    wait_timeout: 5m
spec:
  containers:
    - name: server
      image: aistore/transformer_md5:latest
      imagePullPolicy: IfNotPresent
      ports:
        - name: default
          containerPort: 80
      command: ['/code/server.py', '--listen', '0.0.0.0', '--port', '80']
      readinessProbe:
        httpGet:
          path: /health
          port: default
"""

# returns "Hello World!" on any request.
# pylint: disable=unused-variable
HELLO_WORLD = """
apiVersion: v1
kind: Pod
metadata:
  name: transformer-hello-world
  annotations:
    # Values it can take ["hpull://","hrev://","hpush://"]
    communication_type: "{communication_type}://"
    wait_timeout: 5m
spec:
  containers:
    - name: server
      image: aistore/transformer_hello_world:latest
      imagePullPolicy: IfNotPresent
      ports:
        - name: default
          containerPort: 80
      command: ['/code/server.py']
      # This is a health check endpoint which one should specify
      # for aistore to determine the health of the ETL container.
      readinessProbe:
        httpGet:
          path: /health
          port: default
"""

# returns the original data, with an md5 sum in the response headers.
# pylint: disable=unused-variable
GO_ECHO = """
apiVersion: v1
kind: Pod
metadata:
  name: echo-go
  annotations:
    # Values it can take ["hpull://","hrev://","hpush://"]
    communication_type: "{communication_type}://"
    wait_timeout: 5m
spec:
  containers:
    - name: server
      image: aistore/transformer_echo_go:latest
      imagePullPolicy: IfNotPresent
      ports:
        - name: default
          containerPort: 80
      command: ['./echo', '-l', '0.0.0.0', '-p', '80']
      readinessProbe:
        httpGet:
          path: /health
          port: default
"""

# returns the original data, with an md5 sum in the response headers
# pylint: disable=unused-variable
ECHO = """
apiVersion: v1
kind: Pod
metadata:
  name: transformer-echo
  annotations:
    # Values it can take ["hpull://","hrev://","hpush://"]
    communication_type: "{communication_type}://"
    wait_timeout: 5m
spec:
  containers:
    - name: server
      image: aistore/transformer_echo:latest
      imagePullPolicy: IfNotPresent
      ports:
        - name: default
          containerPort: 80
      command: ['/code/server.py', '--listen', '0.0.0.0', '--port', '80']
      readinessProbe:
        httpGet:
          path: /health
          port: default
"""

# returns the transformed TensorFlow compatible data for the input tar files
# pylint: disable=unused-variable
TAR2TF = """
apiVersion: v1
kind: Pod
metadata:
  name: tar2tf
  annotations:
    # Values it can take ["hpull://","hrev://","hpush://"]
    communication_type: "{communication_type}://"
    wait_timeout: 5m
spec:
  containers:
    - name: server
      image: aistore/transformer_tar2tf:latest
      imagePullPolicy: IfNotPresent
      ports:
        - name: default
          containerPort: 80
      # To enable conversion e.g.
      command: ['./tar2tf', '-l', '0.0.0.0', '-p', '80', '{key}', '{value}']
      readinessProbe:
        httpGet:
          path: /health
          port: default
"""



"""
AIS Plugin for PyTorch

PyTorch Dataset and DataLoader for AIS.

Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
"""

from typing import List, Union
from torch.utils.data import Dataset

from aistore import Client
from aistore.pytorch.utils import list_objects_info

# pylint: disable=too-few-public-methods


class AISBaseClass:
    """
    A base class for creating AIS Datasets for PyTorch

    Args:
        client_url(str): AIS endpoint URL
        urls_list(str or List[str]): single or list of url prefixes to load data
    """

    def __init__(self, client_url: str, urls_list: Union[str, List[str]]) -> None:
        self.client = Client(client_url)
        if isinstance(urls_list, str):
            urls_list = [urls_list]
        self._object_info = {}
        self._object_info = list_objects_info(self.client, urls_list)


# pylint: disable=unused-variable


class AISDataset(AISBaseClass, Dataset):
    """
    A map-style dataset for objects in AIS

    Args:
        client_url(str): AIS endpoint URL
        urls_list(str or List[str]): single or list of url prefixes to load data
    """

    def __init__(self, client_url: str, urls_list: Union[str, List[str]]):
        AISBaseClass.__init__(self, client_url, urls_list)

    def __len__(self):
        return len(self._object_info)

    def __getitem__(self, index: int):
        object_name = self._object_info[index]["object"]
        obj = (
            self.client.bucket(
                bck_name=self._object_info[index]["bck_name"],
                provider=self._object_info[index]["provider"],
            )
            .object(obj_name=object_name)
            .get()
            .read_all()
        )
        return object_name, obj



#
# Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
#

from __future__ import annotations  # pylint: disable=unused-variable
from typing import NewType
import requests

from aistore.client.const import (
    DEFAULT_CHUNK_SIZE,
    HTTP_METHOD_DELETE,
    HTTP_METHOD_GET,
    HTTP_METHOD_HEAD,
    HTTP_METHOD_PUT,
    QParamArchpath,
)

from aistore.client.types import ObjStream

Header = NewType("Header", requests.structures.CaseInsensitiveDict)


# pylint: disable=unused-variable
# pylint: disable=consider-using-with
class Object:
    """
    A class representing an object of a bucket bound to a client.

    Args:
        obj_name (str): name of object
    """

    def __init__(self, bck, obj_name: str):
        self._bck = bck
        self._obj_name = obj_name

    @property
    def bck(self):
        """The custom type [Bck] bound to this object."""
        return self._bck

    @property
    def obj_name(self):
        """The name of this object."""
        return self._obj_name

    def head(self) -> Header:
        """
        Requests object properties.

        Args:
            None

        Returns:
            Response header with the object properties.

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            requests.exeptions.HTTPError(404): The object does not exist
        """
        return self.bck.client.request(
            HTTP_METHOD_HEAD,
            path=f"objects/{ self.bck.name }/{ self.obj_name }",
            params=self.bck.qparam,
        ).headers

    def get(
        self,
        archpath: str = "",
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        etl_id: str = None,
    ) -> ObjStream:
        """
        Reads an object

        Args:
            archpath (str, optional): If the object is an archive, use `archpath` to extract a single file from the archive
            chunk_size (int, optional): chunk_size to use while reading from stream
            etl_id(str, optional): Transforms an object based on ETL with etl_id

        Returns:
            The stream of bytes to read an object or a file inside an archive.

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
        """
        params = self.bck.qparam
        params[QParamArchpath] = archpath
        if etl_id:
            params["uuid"] = etl_id
        resp = self.bck.client.request(
            HTTP_METHOD_GET,
            path=f"objects/{ self.bck.name }/{ self.obj_name }",
            params=params,
            stream=True,
        )
        length = int(resp.headers.get("content-length", 0))
        e_tag = resp.headers.get("ais-checksum-value", "")
        e_tag_type = resp.headers.get("ais-checksum-type", "")
        return ObjStream(
            content_length=length,
            e_tag=e_tag,
            e_tag_type=e_tag_type,
            stream=resp,
            chunk_size=chunk_size,
        )

    def put(self, path: str = None, content: bytes = None) -> Header:
        """
        Puts a local file or bytes as an object to a bucket in AIS storage.

        Args:
            path (str): path to local file or bytes.
            content (bytes): bytes to put as an object.

        Returns:
            Object properties

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            ValueError: Path and content are mutually exclusive
        """
        if path and content:
            raise ValueError("path and content are mutually exclusive")

        url = f"/objects/{ self.bck.name }/{ self.obj_name }"
        if path:
            with open(path, "rb") as reader:
                data = reader.read()
        else:
            data = content
        return self.bck.client.request(
            HTTP_METHOD_PUT,
            path=url,
            params=self.bck.qparam,
            data=data,
        ).headers

    def delete(self):
        """
        Delete an object from a bucket.

        Args:
            None

        Returns:
            None

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            requests.exeptions.HTTPError(404): The object does not exist
        """
        self.bck.client.request(
            HTTP_METHOD_DELETE,
            path=f"objects/{ self.bck.name }/{ self.obj_name }",
            params=self.bck.qparam,
        )



#
# Copyright (c) 2018-2022, NVIDIA CORPORATION. All rights reserved.
#

from __future__ import annotations  # pylint: disable=unused-variable
from typing import TypeVar, Type
import requests
from urllib.parse import urljoin
from pydantic.tools import parse_raw_as

from aistore.client.bucket import Bucket
from aistore.client.const import (
    HTTP_METHOD_GET,
    ProviderAIS,
    QParamArchpath,
    QParamProvider,
)
from aistore.client.cluster import Cluster
from aistore.client.types import BucketLister, ObjStream, Namespace
from aistore.client.utils import handle_errors
from aistore.client.xaction import Xaction
from aistore.client.etl import Etl

T = TypeVar("T")


# pylint: disable=unused-variable
class Client:
    """
    AIStore client for managing buckets, objects, ETL jobs

    Args:
        endpoint (str): AIStore endpoint
    """

    def __init__(self, endpoint: str):
        self._endpoint = endpoint
        self._base_url = urljoin(self._endpoint, "v1")
        self._session = requests.session()

    @property
    def base_url(self):
        return self._base_url

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def session(self):
        return self._session

    def request_deserialize(
        self, method: str, path: str, res_model: Type[T], **kwargs
    ) -> T:
        resp = self.request(method, path, **kwargs)
        return parse_raw_as(res_model, resp.text)

    def request(self, method: str, path: str, **kwargs) -> requests.Response:
        url = f"{ self.base_url }/{ path.lstrip('/') }"
        resp = self.session.request(
            method, url, headers={"Accept": "application/json"}, **kwargs
        )
        if resp.status_code < 200 or resp.status_code >= 300:
            handle_errors(resp)
        return resp

    def bucket(self, bck_name: str, provider: str = ProviderAIS, ns: Namespace = None):
        """
        Factory constructor for bucket object.
        Does not make any HTTP request, only instantiates a bucket object owned by the client.

        Args:
            bck_name (str): Name of bucket (optional, defaults to "ais").
            provider (str): Provider of bucket (one of "ais", "aws", "gcp", ...).
            ns (Namespace): Namespace of bucket (optional, defaults to None).

        Returns:
            The bucket object created.
        """
        return Bucket(client=self, bck_name=bck_name, provider=provider, ns=ns)

    def cluster(self):
        """
        Factory constructor for cluster object.
        Does not make any HTTP request, only instantiates a cluster object owned by the client.

        Args:
            None

        Returns:
            The cluster object created.
        """
        return Cluster(client=self)

    def xaction(self):
        """
        Factory constructor for xaction object, which contains xaction-related functions.
        Does not make any HTTP request, only instantiates an xaction object bound to the client.

        Args:
            None

        Returns:
            The xaction object created.
        """
        return Xaction(client=self)

    def etl(self):
        """
        Factory constructor for ETL object.
        Contains APIs related to AIStore ETL operations.
        Does not make any HTTP request, only instantiates an xaction object bound to the client.

        Args:
            None

        Returns:
            The xaction object created.
        """
        return Etl(client=self)

    # TODO: Remove once pytorch/data dependency on previous version is resolved
    def list_objects_iter(
        self,
        bck_name: str,
        provider: str = ProviderAIS,
        prefix: str = "",
        props: str = "",
        page_size: int = 0,
    ) -> BucketLister:
        """
        Returns an iterator for all objects in a bucket
        Args:
            bck_name (str): Name of a bucket
            provider (str, optional): Name of bucket provider, one of "ais", "aws", "gcp", "az", "hdfs" or "ht".
                Defaults to "ais". Empty provider returns buckets of all providers.
            prefix (str, optional): return only objects that start with the prefix
            props (str, optional): comma-separated list of object properties to return. Default value is "name,size". Properties: "name", "size", "atime", "version", "checksum", "cached", "target_url", "status", "copies", "ec", "custom", "node".
        Returns:
            BucketLister: object iterator
        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
        """
        return BucketLister(
            self,
            bck_name=bck_name,
            provider=provider,
            prefix=prefix,
            props=props,
            page_size=page_size,
        )

    # TODO: Remove once pytorch/data dependency on previous version is resolved
    def get_object(
        self,
        bck_name: str,
        obj_name: str,
        provider: str = ProviderAIS,
        archpath: str = "",
        chunk_size: int = 1,
    ) -> ObjStream:
        """
        Reads an object
        Args:
            bck_name (str): Name of a bucket
            obj_name (str): Name of an object in the bucket
            provider (str, optional): Name of bucket provider, one of "ais", "aws", "gcp", "az", "hdfs" or "ht".
            archpath (str, optional): If the object is an archive, use `archpath` to extract a single file from the archive
            chunk_size (int, optional): chunk_size to use while reading from stream
        Returns:
            The stream of bytes to read an object or a file inside an archive.
        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
        """
        params = {QParamProvider: provider, QParamArchpath: archpath}
        resp = self.request(
            HTTP_METHOD_GET,
            path=f"objects/{ bck_name }/{ obj_name }",
            params=params,
            stream=True,
        )
        length = int(resp.headers.get("content-length", 0))
        e_tag = resp.headers.get("ais-checksum-value", "")
        e_tag_type = resp.headers.get("ais-checksum-type", "")
        return ObjStream(
            content_length=length,
            e_tag=e_tag,
            e_tag_type=e_tag_type,
            stream=resp,
            chunk_size=chunk_size,
        )



#
# Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
#

import base64
from typing import Callable, List
import cloudpickle
from aistore.client.const import (
    CODE_TEMPLATE,
    HTTP_METHOD_DELETE,
    HTTP_METHOD_GET,
    HTTP_METHOD_POST,
    HTTP_METHOD_PUT,
)
from aistore.client.types import ETL, ETLDetails

# pylint: disable=unused-variable
class Etl:
    """
    A class containing ETL-related functions.

    Args:
        None
    """

    def __init__(self, client):
        self._client = client

    @property
    def client(self):
        """The client bound to this ETL object."""
        return self._client

    def init_spec(
        self,
        template: str,
        etl_id: str,
        communication_type: str = "hpush",
        timeout: str = "5m",
    ):
        """
        Initializes ETL based on POD spec template. Returns ETL_ID.
        Existing templates can be found at `aistore.client.etl_templates`
        For more information visit: https://github.com/NVIDIA/ais-etl/tree/master/transformers

        Args:
            docker_image (str): docker image name looks like: <hub-user>/<repo-name>:<tag>
            etl_id (str): id of new ETL
            communication_type (str): Communication type of the ETL (options: hpull, hrev, hpush)
            timeout (str): timeout of the ETL (eg. 5m for 5 minutes)
        Returns:
            etl_id (str): ETL ID
        """

        # spec
        spec_encoded = base64.b64encode(template.encode("utf-8")).decode("utf-8")

        action = {
            "spec": spec_encoded,
            "id": etl_id,
            "communication": f"{communication_type}://",
            "timeout": timeout,
        }

        resp = self.client.request(HTTP_METHOD_PUT, path="etl", json=action)
        return resp.text

    def init_code(
        self,
        transform: Callable,
        etl_id: str,
        dependencies: List[str] = None,
        runtime: str = "python3.8v2",
        communication_type: str = "hpush",
        timeout: str = "5m",
        chunk_size: int = None,
    ):
        """
        Initializes ETL based on the provided source code. Returns ETL_ID.

        Args:
            transform (Callable): Transform function of the ETL
            etl_id (str): Id of new ETL
            runtime (str): [optional, default="python3.8v2"] Runtime environment of the ETL [choose from: python3.8v2, python3.10v2] (see etl/runtime/all.go)
            communication_type (str): [optional, default="hpush"] Communication type of the ETL (options: hpull, hrev, hpush, io)
            timeout (str): [optional, default="5m"] Timeout of the ETL (eg. 5m for 5 minutes)
            chunk_size (int): Chunk size in bytes if transform function in streaming data. (whole object is read by default)
        Returns:
            etl_id (str): ETL ID
        """
        if communication_type not in ["io", "hpush", "hrev", "hpull"]:
            raise ValueError("communication_type should be in: hpull, hrev, hpush, io")

        functions = {
            "transform": "transform",
        }

        action = {
            "id": etl_id,
            "runtime": runtime,
            "communication": f"{communication_type}://",
            "timeout": timeout,
            "funcs": functions,
        }

        if chunk_size:
            action["chunk_size"] = chunk_size

        # code
        transform = base64.b64encode(cloudpickle.dumps(transform)).decode("utf-8")

        io_comm_context = "transform()" if communication_type == "io" else ""
        template = CODE_TEMPLATE.format(transform, io_comm_context).encode("utf-8")
        action["code"] = base64.b64encode(template).decode("utf-8")

        # dependencies
        if dependencies is None:
            dependencies = []
        dependencies.append("cloudpickle==2.0.0")
        deps = "\n".join(dependencies).encode("utf-8")
        action["dependencies"] = base64.b64encode(deps).decode("utf-8")

        resp = self.client.request(
            HTTP_METHOD_PUT,
            path="etl",
            json=action,
        )
        return resp.text

    def list(self) -> List[ETLDetails]:
        """
        Lists all running ETLs.

        Note: Does not list ETLs that have been stopped or deleted.

        Args:
            Nothing
        Returns:
            List[ETL]: A list of running ETLs
        """
        resp = self.client.request_deserialize(
            HTTP_METHOD_GET, path="etl", res_model=List[ETL]
        )
        return resp

    def view(self, etl_id: str) -> ETLDetails:
        """
        View ETLs Init spec/code

        Args:
            etl_id (str): id of ETL
        Returns:
            ETLDetails: details of the ETL
        """
        resp = self.client.request_deserialize(
            HTTP_METHOD_GET, path=f"etl/{ etl_id }", res_model=ETLDetails
        )
        return resp

    def start(self, etl_id: str):
        """
        Resumes a stopped ETL with given ETL_ID.

        Note: Deleted ETLs cannot be started.

        Args:
            etl_id (str): id of ETL
        Returns:
            Nothing
        """
        self.client.request(HTTP_METHOD_POST, path=f"etl/{ etl_id }/start")

    def stop(self, etl_id: str):
        """
        Stops ETL with given ETL_ID. Stops (but does not delete) all the pods created by Kubernetes for this ETL and terminates any transforms.

        Args:
            etl_id (str): id of ETL
        Returns:
            Nothing
        """
        self.client.request(HTTP_METHOD_POST, path=f"etl/{ etl_id }/stop")

    def delete(self, etl_id: str):
        """
        Delete ETL with given ETL_ID. Deletes pods created by Kubernetes for this ETL and specifications for this ETL in Kubernetes.

        Note: Running ETLs cannot be deleted.

        Args:
            etl_id (str): id of ETL
        Returns:
            Nothing
        """
        self.client.request(HTTP_METHOD_DELETE, path=f"etl/{ etl_id }")



#
# Copyright (c) 2021-22, NVIDIA CORPORATION. All rights reserved.
#

# pylint: disable=unused-variable

TAR2TF = "tar2tf"
OBJECTS = "objects"
START = "start"

# URL Params
QParamArchpath = "archpath"
QParamProvider = "provider"
QParamWhat = "what"
QParamKeepBckMD = "keep_md"
QParamBucketTo = "bck_to"
QparamPrimaryReadyReb = "prr"

# Bucket providers
ProviderAIS = "ais"
ProviderAmazon = "aws"
ProviderGoogle = "gcp"
ProviderHTTP = "ht"
ProviderAzure = "azure"
ProviderHDFS = "hdfs"

HTTP_METHOD_GET = "get"
HTTP_METHOD_POST = "post"
HTTP_METHOD_DELETE = "delete"
HTTP_METHOD_PUT = "put"
HTTP_METHOD_HEAD = "head"

# Actions
ACT_COPY_BCK = "copy-bck"
ACT_CREATE_BCK = "create-bck"
ACT_DESTROY_BCK = "destroy-bck"
ACT_EVICT_REMOTE_BCK = "evict-remote-bck"
ACT_LIST = "list"
ACT_MOVE_BCK = "move-bck"
ACT_ETL_BCK = "etl-bck"

# Defaults
DEFAULT_CHUNK_SIZE = 32768

# templates for ETL

CODE_TEMPLATE = """
import pickle
import base64

transform = pickle.loads(base64.b64decode('{}'))
{}
"""



"""
Utils for AIS PyTorch Plugin

Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
"""

from typing import List, Mapping, Tuple
from urllib.parse import urlparse, urlunparse
from aistore import Client


def parse_url(url: str) -> Tuple[str, str, str]:
    """
    Parse AIS urls for bucket and object names
    Args:
        url (str): Complete URL of the object (eg. "ais://bucket1/file.txt")
    Returns:
        provider (str): AIS Backend
        bck_name (str): Bucket name identifier
        obj_name (str):  Object name with extension
    """
    parsed_url = urlparse(url)
    path = parsed_url.path
    if len(path) > 0 and path.startswith("/"):
        path = path[1:]

    # returns provider, bck_name, path
    return parsed_url.scheme, parsed_url.netloc, path


# pylint: disable=unused-variable
def list_objects_info(client: Client, urls_list: List[str]) -> List[Mapping[str, str]]:
    """
    Create list of list of [bucket_name, object_name] from all the object urls
    Args:
        client (Client): AIStore client object of the calling method
        urls_list (List[str]): list of urls
    Returns:
        List[samples](List[Mapping[str, str]]): list of {provider, bucket, path to the object}
    """
    samples = []
    for url in urls_list:
        provider, bck_name, path = parse_url(url)
        objects = client.bucket(bck_name=bck_name, provider=provider).list_objects(
            prefix=path
        )
        for obj_info in objects.get_entries():
            samples.append(
                {"provider": provider, "bck_name": bck_name, "object": obj_info.name}
            )
    return samples


def unparse_url(provider: str, bck_name: str, obj_name: str) -> str:
    """
    To generate URL based on provider, bck_name and object name
    Args:
        provider(str): Provider name ('ais', 'gcp', etc)
        bck_name(str): Bucket name
        obj_name(str): Object name with extension.
    Returns:
        unparsed_url(str): Unparsed url (complete url)
    """
    return urlunparse([provider, bck_name, obj_name, "", "", ""])



#
# Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
#

from __future__ import annotations  # pylint: disable=unused-variable
from typing import Dict, List, NewType
import requests

from aistore.client.const import (
    ACT_COPY_BCK,
    ACT_CREATE_BCK,
    ACT_DESTROY_BCK,
    ACT_ETL_BCK,
    ACT_EVICT_REMOTE_BCK,
    ACT_LIST,
    ACT_MOVE_BCK,
    HTTP_METHOD_DELETE,
    HTTP_METHOD_GET,
    HTTP_METHOD_HEAD,
    HTTP_METHOD_POST,
    ProviderAIS,
    QParamBucketTo,
    QParamKeepBckMD,
    QParamProvider,
)

from aistore.client.errors import InvalidBckProvider
from aistore.client.object import Object
from aistore.client.types import (
    ActionMsg,
    Bck,
    BucketEntry,
    BucketList,
    BucketLister,
    Namespace,
)

Header = NewType("Header", requests.structures.CaseInsensitiveDict)


# pylint: disable=unused-variable
class Bucket:
    """
    A class representing a bucket that contains user data.

    Args:
        bck_name (str): name of bucket
        provider (str, optional): provider of bucket (one of "ais", "aws", "gcp", ...), defaults to "ais"
        ns (Namespace, optional): namespace of bucket, defaults to None
    """

    def __init__(
        self, client, bck_name: str, provider: str = ProviderAIS, ns: Namespace = None
    ):
        self._client = client
        self._bck = Bck(name=bck_name, provider=provider, ns=ns)
        self._qparam = {QParamProvider: provider}

    @property
    def client(self):
        """The client bound to this bucket."""
        return self._client

    @property
    def bck(self):
        """The custom type [Bck] corresponding to this bucket."""
        return self._bck

    @property
    def qparam(self):
        """The QParamProvider of this bucket."""
        return self._qparam

    @property
    def provider(self):
        """The provider for this bucket."""
        return self.bck.provider

    @property
    def name(self):
        """The name of this bucket."""
        return self.bck.name

    @property
    def namespace(self):
        """The namespace for this bucket."""
        return self.bck.ns

    def create(self):
        """
        Creates a bucket in AIStore cluster.
        Can only create a bucket for AIS provider on localized cluster. Remote cloud buckets do not support creation.
        Args:
            None

        Returns:
            None

        Raises:
            aistore.client.errors.AISError: All other types of errors with AIStore
            aistore.client.errors.InvalidBckProvider: Invalid bucket provider for requested operation
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ReadTimeout: Timed out receiving response from AIStore
        """
        if self.provider is not ProviderAIS:
            raise InvalidBckProvider(self.provider)
        action = ActionMsg(action=ACT_CREATE_BCK).dict()
        self.client.request(
            HTTP_METHOD_POST,
            path=f"buckets/{ self.name }",
            json=action,
            params=self.qparam,
        )

    def delete(self):
        """
        Destroys bucket in AIStore cluster.
        In all cases removes both the bucket's content _and_ the bucket's metadata from the cluster.
        Note: AIS will _not_ call the remote backend provider to delete the corresponding Cloud bucket
        (iff the bucket in question is, in fact, a Cloud bucket).

        Args:
            None

        Returns:
            None

        Raises:
            aistore.client.errors.AISError: All other types of errors with AIStore
            aistore.client.errors.InvalidBckProvider: Invalid bucket provider for requested operation
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ReadTimeout: Timed out receiving response from AIStore
        """
        if self.provider is not ProviderAIS:
            raise InvalidBckProvider(self.provider)
        action = ActionMsg(action=ACT_DESTROY_BCK).dict()
        self.client.request(
            HTTP_METHOD_DELETE,
            path=f"buckets/{ self.name }",
            json=action,
            params=self.qparam,
        )

    def rename(self, to_bck: str) -> str:
        """
        Renames bucket in AIStore cluster.
        Only works on AIS buckets. Returns xaction id that can be used later to check the status of the asynchronous operation.

        Args:
            to_bck (str): New bucket name for bucket to be renamed as

        Returns:
            xaction id (as str) that can be used to check the status of the operation

        Raises:
            aistore.client.errors.AISError: All other types of errors with AIStore
            aistore.client.errors.InvalidBckProvider: Invalid bucket provider for requested operation
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ReadTimeout: Timed out receiving response from AIStore
        """
        if self.provider is not ProviderAIS:
            raise InvalidBckProvider(self.provider)
        params = self.qparam.copy()
        params[QParamBucketTo] = f"{ProviderAIS}/@#/{to_bck}/"
        action = ActionMsg(action=ACT_MOVE_BCK).dict()
        resp = self.client.request(
            HTTP_METHOD_POST, path=f"buckets/{ self.name }", json=action, params=params
        )
        self.bck.name = to_bck
        return resp.text

    def evict(self, keep_md: bool = False):
        """
        Evicts bucket in AIStore cluster.
        NOTE: only Cloud buckets can be evicted.

        Args:
            keep_md (bool, optional): If true, evicts objects but keeps the bucket's metadata (i.e., the bucket's name and its properties)

        Returns:
            None

        Raises:
            aistore.client.errors.AISError: All other types of errors with AIStore
            aistore.client.errors.InvalidBckProvider: Invalid bucket provider for requested operation
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ReadTimeout: Timed out receiving response from AIStore
        """
        if self.provider is ProviderAIS:
            raise InvalidBckProvider(self.provider)
        params = self.qparam.copy()
        params[QParamKeepBckMD] = keep_md
        action = ActionMsg(action=ACT_EVICT_REMOTE_BCK).dict()
        self.client.request(
            HTTP_METHOD_DELETE,
            path=f"buckets/{ self.name }",
            json=action,
            params=params,
        )

    def head(self) -> Header:
        """
        Requests bucket properties.

        Args:
            None

        Returns:
            Response header with the bucket properties

        Raises:
            aistore.client.errors.AISError: All other types of errors with AIStore
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ReadTimeout: Timed out receiving response from AIStore
        """
        return self.client.request(
            HTTP_METHOD_HEAD,
            path=f"buckets/{ self.name }",
            params=self.qparam,
        ).headers

    def copy(
        self,
        to_bck_name: str,
        prefix: str = "",
        dry_run: bool = False,
        force: bool = False,
        to_provider: str = ProviderAIS,
    ) -> str:
        """
        Returns xaction id that can be used later to check the status of the asynchronous operation.

        Args:
            to_bck_name (str): Name of the destination bucket
            prefix (str, optional): If set, only the objects starting with
                provider prefix will be copied
            dry_run (bool, optional): Determines if the copy should actually
                happen or not
            force (bool, optional): Override existing destination bucket
            to_provider (str, optional): Name of destination bucket provider

        Returns:
            Xaction id (as str) that can be used to check the status of the operation

        Raises:
            aistore.client.errors.AISError: All other types of errors with AIStore
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ReadTimeout: Timed out receiving response from AIStore
        """

        value = {"prefix": prefix, "dry_run": dry_run, "force": force}
        action = ActionMsg(action=ACT_COPY_BCK, value=value).dict()
        params = self.qparam.copy()
        params[QParamBucketTo] = f"{ to_provider }/@#/{ to_bck_name }/"
        return self.client.request(
            HTTP_METHOD_POST,
            path=f"buckets/{ self.name }",
            json=action,
            params=params,
        ).text

    def list_objects(
        self,
        prefix: str = "",
        props: str = "",
        page_size: int = 0,
        uuid: str = "",
        continuation_token: str = "",
    ) -> BucketList:
        """
        Returns a structure that contains a page of objects, xaction UUID, and continuation token (to read the next page, if available).

        Args:
            prefix (str, optional): Return only objects that start with the prefix
            props (str, optional): Comma-separated list of object properties to return. Default value is "name,size". Properties: "name", "size", "atime", "version", "checksum", "cached", "target_url", "status", "copies", "ec", "custom", "node".
            page_size (int, optional): Return at most "page_size" objects.
                The maximum number of objects in response depends on the bucket backend. E.g, AWS bucket cannot return more than 5,000 objects in a single page.
                NOTE: If "page_size" is greater than a backend maximum, the backend maximum objects are returned.
                Defaults to "0" - return maximum number objects.
            uuid (str, optional): Job UUID, required to get the next page of objects
            continuation_token (str, optional): Marks the object to start reading the next page

        Returns:
            BucketList: the page of objects in the bucket and the continuation token to get the next page
            Empty continuation token marks the final page of the object list

        Raises:
            aistore.client.errors.AISError: All other types of errors with AIStore
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ReadTimeout: Timed out receiving response from AIStore
        """
        value = {
            "prefix": prefix,
            "pagesize": page_size,
            "uuid": uuid,
            "props": props,
            "continuation_token": continuation_token,
        }
        action = ActionMsg(action=ACT_LIST, value=value).dict()

        return self.client.request_deserialize(
            HTTP_METHOD_GET,
            path=f"buckets/{ self.name }",
            res_model=BucketList,
            json=action,
            params=self.qparam,
        )

    def list_objects_iter(
        self,
        prefix: str = "",
        props: str = "",
        page_size: int = 0,
    ) -> BucketLister:
        """
        Returns an iterator for all objects in bucket

        Args:
            prefix (str, optional): Return only objects that start with the prefix
            props (str, optional): Comma-separated list of object properties to return. Default value is "name,size". Properties: "name", "size", "atime", "version", "checksum", "cached", "target_url", "status", "copies", "ec", "custom", "node".
            page_size (int, optional): return at most "page_size" objects
                The maximum number of objects in response depends on the bucket backend. E.g, AWS bucket cannot return more than 5,000 objects in a single page.
                NOTE: If "page_size" is greater than a backend maximum, the backend maximum objects are returned.
                Defaults to "0" - return maximum number objects

        Returns:
            BucketLister: object iterator

        Raises:
            aistore.client.errors.AISError: All other types of errors with AIStore
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ReadTimeout: Timed out receiving response from AIStore
        """
        return BucketLister(
            self.client,
            bck_name=self.name,
            provider=self.provider,
            prefix=prefix,
            props=props,
            page_size=page_size,
        )

    def list_all_objects(
        self,
        prefix: str = "",
        props: str = "",
        page_size: int = 0,
    ) -> List[BucketEntry]:
        """
        Returns a list of all objects in bucket

        Args:
            prefix (str, optional): return only objects that start with the prefix
            props (str, optional): comma-separated list of object properties to return. Default value is "name,size". Properties: "name", "size", "atime", "version", "checksum", "cached", "target_url", "status", "copies", "ec", "custom", "node".
            page_size (int, optional): return at most "page_size" objects
                The maximum number of objects in response depends on the bucket backend. E.g, AWS bucket cannot return more than 5,000 objects in a single page.
                NOTE: If "page_size" is greater than a backend maximum, the backend maximum objects are returned.
                Defaults to "0" - return maximum number objects

        Returns:
            List[BucketEntry]: list of objects in bucket

        Raises:
            aistore.client.errors.AISError: All other types of errors with AIStore
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ReadTimeout: Timed out receiving response from AIStore
        """
        uuid = ""
        continuation_token = ""
        obj_list = None

        while True:
            resp = self.list_objects(
                prefix=prefix,
                props=props,
                page_size=page_size,
                uuid=uuid,
                continuation_token=continuation_token,
            )
            if obj_list:
                obj_list = obj_list + resp.get_entries()
            obj_list = obj_list or resp.get_entries()
            if resp.continuation_token == "":
                break
            continuation_token = resp.continuation_token
            uuid = resp.uuid
        return obj_list

    def transform(
        self,
        etl_id: str,
        to_bck: str,
        prefix: str = "",
        ext: Dict[str, str] = None,
        force: bool = False,
        dry_run: bool = False,
    ):
        """
        Transforms all objects in a bucket and puts them to destination bucket.

        Args:
            etl_id (str): id of etl to be used for transformations
            to_bck (str): destination bucket for transformations
            prefix (str, optional): prefix to be added to resulting transformed objects
            ext (Dict[str, str], optional): dict of new extension followed by extension to be replaced (i.e. {"jpg": "txt"})
            dry_run (bool, optional): determines if the copy should actually happen or not
            force (bool, optional): override existing destination bucket

        Returns:
            Xaction id (as str) that can be used to check the status of the operation
        """
        value = {
            "id": etl_id,
            "prefix": prefix,
            "force": force,
            "dry_run": dry_run,
        }

        if ext:
            value["ext"] = ext

        action = ActionMsg(action=ACT_ETL_BCK, value=value).dict()

        params = self.qparam.copy()
        params[QParamBucketTo] = f"{ProviderAIS}/@#/{to_bck}/"

        resp = self.client.request(
            HTTP_METHOD_POST,
            path=f"buckets/{ self.name }",
            json=action,
            params=params,
        )

        return resp.text

    def object(self, obj_name: str):
        """
        Factory constructor for object bound to bucket.
        Does not make any HTTP request, only instantiates an object in a bucket owned by the client.

        Args:
            obj_name (str): Name of object

        Returns:
            The object created.
        """
        return Object(bck=self, obj_name=obj_name)



#!/usr/bin/env python3

import os
import sys
from pathlib import Path

from setuptools import setup, find_packages

REQUIRED_MAJOR = 3
REQUIRED_MINOR = 6

ROOT_DIR = Path(__file__).parent.resolve()

VERSION = "1.0.4"


def _get_version() -> str:
    # TODO: read from file
    # with open(os.path.join(ROOT_DIR, "version.txt"), encoding='UTF-8') as file:
    #     version = file.readline().strip()
    # return version
    return VERSION


def _export_version(version):
    version_path = ROOT_DIR / "aistore" / "version.py"
    with open(version_path, "w", encoding="UTF-8") as file:
        file.write(f"__version__ = '{version}'\n")


# Check for python version
if sys.version_info < (REQUIRED_MAJOR, REQUIRED_MINOR):
    error = (
        "Your version of python ({major}.{minor}) is too old. You need "
        "python >= {required_major}.{required_minor}."
    ).format(
        major=sys.version_info.major,
        minor=sys.version_info.minor,
        required_minor=REQUIRED_MINOR,
        required_major=REQUIRED_MAJOR,
    )
    sys.exit(error)

# Read in README.md for our long_description
with open(os.path.join(ROOT_DIR, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

VERSION = _get_version()
_export_version(VERSION)

setup(
    name="aistore",
    version=VERSION,
    description="A (growing) set of client-side APIs to access and utilize clusters, buckets, and objects on AIStore.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://aiatscale.org",
    download_url="https://github.com/NVIDIA/aistore/tags",
    author="AIStore Team",
    author_email="aistore@exchange.nvidia.com",
    keywords=[
        "AIStore",
        "Artificial Intelligence",
        "Object Storage",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Scientific/Engineering",
    ],
    project_urls={
        "Documentation": "https://aiatscale.org/docs/",
        "Release notes": "https://github.com/NVIDIA/aistore/releases/",
        "Source": "https://github.com/NVIDIA/aistore/",
    },
    license="MIT",
    python_requires=">=3.6",
    packages=find_packages(exclude=("tests", "tests.*", "examples", "examples.*")),
    # packaging used to check latest version in torchdata
    install_requires=["requests", "packaging", "pydantic==1.9.0", "cloudpickle==2.0.0"],
    extras_require={
        "pytorch": ["torch", "torchdata"],
        "botocore": ["wrapt"],
    },
)



#
# Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
#
"""
Allow use of Amazon's boto3 and botocore libraries with AIStore.

If you would like to interact with AIStore using the popular
boto3 (or botocore) libraries, you can do so by importing this
module (you'll also need to install wrapt):

```
import wrapt
from aistore.monkey import botocore
```

By way of explanation: by default, botocore (used by boto3)
does not respect HTTP redirects that make use of the
`location` header. Since AIStore uses them for load
balancing, trying to use boto without patching it first
results in various ClientErrors.

Importing this module patches the botocore library to respect
the location header under the same conditions used by AIStore:
the response code is 301, 302 or 307, and the "location" header
is set.
"""

import inspect
import wrapt


# pylint: disable=unused-argument
def _ais_redirect_wrapper(wrapped, instance, args, kwargs):
    """
    Intercept S3 redirections from AIStore and patch them to:

      - Respect HTTP 301, 302 or 307 redirects using the 'Location' header
      - Fail after 3 redirects for a given operation (rather than just one)

    We make no attempt to differentiate semantically between different 3XX
    codes; we just reuse the same method and request context, and hit the
    url specified by the "location" response header.

    We also don't try to handle caching via 'Cache-Control' or 'Expires'
    (though we might, later).
    """
    response = kwargs.get("response")
    request_dict = kwargs.get("request_dict")

    intercept_criteria = [
        lambda: response is not None,
        lambda: isinstance(response, tuple) and isinstance(request_dict, dict),
        lambda: response[0].status_code in [301, 302, 307],
        lambda: request_dict["context"]["s3_redirect"].get("ais_redirect_count", 0)
        <= 3,
        lambda: {k.lower(): v for k, v in response[0].headers.items()}.get("location"),
    ]

    try:
        if not all(i() for i in intercept_criteria):
            return wrapped(*args, **kwargs)
    except Exception:  # pylint: disable=broad-except
        return wrapped(*args, **kwargs)

    if request_dict["context"]["s3_redirect"].get("ais_redirect_count") is None:
        request_dict["context"]["s3_redirect"]["ais_redirect_count"] = 0
    request_dict["context"]["s3_redirect"]["ais_redirect_count"] += 1

    request_dict["url"] = response[0].headers["location"]

    return 0


@wrapt.when_imported("botocore.utils")
def _apply_patches(module):  # pylint: disable=unused-variable
    """
    When botocore.utils is imported, patch it to handle redirects.

    We can't future proof against everything, but we make a reasonable
    effort to find all likely "S3RegionRedirector" type classes below
    (at time of writing there are two, "S3RegionRedirector" and "S3RegionRedirectorv2".

    If the S3RegionRedirector.redirect_on_error method bound to each
    fails to match our signature, we'll bail.
    """
    redirectors = [
        c[1]
        for c in inspect.getmembers(module, inspect.isclass)
        if "S3RegionRedirector" in c[0]
    ]

    def _expected_sig(self, request_dict, response, operation, **kwargs):
        pass

    for redirector in redirectors:
        assert inspect.signature(redirector.redirect_from_error) == inspect.signature(
            _expected_sig
        )
        wrapt.wrap_function_wrapper(
            module, redirector.__name__ + ".redirect_from_error", _ais_redirect_wrapper
        )



#
# Copyright (c) 2021-2022, NVIDIA CORPORATION. All rights reserved.
#

from __future__ import annotations  # pylint: disable=unused-variable
import base64

from typing import Any, Mapping, List, Iterator, Optional

from pydantic import BaseModel, Field, StrictInt, StrictStr, validator
import requests
from aistore.client.const import DEFAULT_CHUNK_SIZE, ProviderAIS

# pylint: disable=too-few-public-methods,unused-variable


class Namespace(BaseModel):
    uuid: str = ""
    name: str = ""


class Bck(BaseModel):
    name: str
    provider: str = ProviderAIS
    ns: Namespace = None


class ActionMsg(BaseModel):
    action: str
    name: str = ""
    value: Any = None


class NetInfo(BaseModel):
    node_hostname: str = ""
    daemon_port: str = ""
    direct_url: str = ""


class Snode(BaseModel):
    daemon_id: str
    daemon_type: str
    public_net: NetInfo = None
    intra_control_net: NetInfo = None
    intra_data_net: NetInfo = None
    flags: int = 0


class Smap(BaseModel):
    tmap: Mapping[str, Snode]
    pmap: Mapping[str, Snode]
    proxy_si: Snode
    version: int = 0
    uuid: str = ""
    creation_time: str = ""


class BucketEntry(BaseModel):
    name: str
    size: int = 0
    checksum: str = ""
    atime: str = ""
    version: str = ""
    target_url: str = ""
    copies: int = 0
    flags: int = 0

    def is_cached(self):
        return (self.flags & (1 << 6)) != 0

    def is_ok(self):
        return (self.flags & ((1 << 5) - 1)) == 0


class BucketList(BaseModel):
    uuid: str
    entries: Optional[List[BucketEntry]] = []
    continuation_token: str
    flags: int

    def get_entries(self):
        return self.entries

    @validator("entries")
    def set_entries(cls, entries):  # pylint: disable=no-self-argument
        if entries is None:
            entries = []
        return entries


class XactStatus(BaseModel):
    uuid: str = ""
    err: str = ""
    end_time: int = 0
    aborted: bool = False


class ObjStream(BaseModel):
    class Config:
        validate_assignment = True
        arbitrary_types_allowed = True

    content_length: StrictInt = Field(default=-1, allow_mutation=False)
    chunk_size: StrictInt = Field(default=DEFAULT_CHUNK_SIZE, allow_mutation=False)
    e_tag: StrictStr = Field(..., allow_mutation=False)
    e_tag_type: StrictStr = Field(..., allow_mutation=False)
    stream: requests.Response

    # read_all uses a bytes cast which makes it slightly slower
    def read_all(self) -> bytes:
        obj_arr = bytearray()
        for chunk in self:
            obj_arr.extend(chunk)
        return bytes(obj_arr)

    def raw(self) -> bytes:
        return self.stream.raw

    def __iter__(self) -> Iterator[bytes]:
        try:
            for chunk in self.stream.iter_content(chunk_size=self.chunk_size):
                yield chunk
        finally:
            self.stream.close()


class HttpError(BaseModel):
    status: int
    message: str = ""
    method: str = ""
    url_path: str = ""
    remote_addr: str = ""
    caller: str = ""
    node: str = ""


class BucketLister:
    _fetched: Optional[List[BucketEntry]] = []
    _token: str = ""
    _uuid: str = ""
    _prefix: str = ""
    _props: str = ""
    _provider: str = ProviderAIS
    _bck_name: str = ""
    _client: "Client"

    def __init__(
        self,
        client: "Client",
        bck_name: str = "",
        provider: str = ProviderAIS,
        prefix: str = "",
        props: str = "",
        page_size: int = 0,
    ):
        self._client = client
        self._prefix = prefix
        self._props = props
        self._bck_name = bck_name
        self._provider = provider
        self._page_size = page_size

    def __iter__(self):
        return self

    def __next__(self) -> Iterator[BucketEntry]:
        # Iterator is exhausted.
        if len(self._fetched) == 0 and self._token == "" and self._uuid != "":
            raise StopIteration
        # Read the next page of objects.
        if len(self._fetched) == 0:
            value = {
                "prefix": self._prefix,
                "uuid": self._uuid,
                "props": self._props,
                "continuation_token": self._token,
                "pagesize": self._page_size,
            }
            resp = self._client.bucket(self._bck_name, self._provider).list_objects(
                prefix=self._prefix,
                props=self._props,
                uuid=self._uuid,
                continuation_token=self._token,
                page_size=self._page_size,
            )
            self._fetched = resp.get_entries()
            self._uuid = resp.uuid
            self._token = resp.continuation_token
            # Empty page and token mean no more objects left.
            if len(self._fetched) == 0 and self._token == "":
                raise StopIteration
        return self._fetched.pop(0)


class ETL(BaseModel):  # pylint: disable=too-few-public-methods,unused-variable
    id: str = ""
    obj_count: int = 0
    in_bytes: int = 0
    out_bytes: int = 0


class ETLDetails(BaseModel):
    id: str
    communication: str
    timeout: str
    code: Optional[bytes]
    spec: Optional[str]
    dependencies: Optional[str]
    runtime: str = "python3.8v2"  # see etl/runtime/all.go
    chunk_size: int = 0

    @validator("code")
    def set_code(cls, code):  # pylint: disable=no-self-argument
        if code is not None:
            code = base64.b64decode(code)
        return code

    @validator("spec")
    def set_spec(cls, spec):  # pylint: disable=no-self-argument
        if spec is not None:
            spec = base64.b64decode(spec)
        return spec



#
# Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
#

from __future__ import annotations  # pylint: disable=unused-variable
from typing import List

from aistore.client.const import (
    HTTP_METHOD_GET,
    ACT_LIST,
    ProviderAIS,
    QParamWhat,
    QparamPrimaryReadyReb,
    QParamProvider,
)

from aistore.client.types import ActionMsg, Bck, Smap


# pylint: disable=unused-variable
class Cluster:
    """
    A class representing a cluster bound to an AIS client.

    Args:
        None
    """

    def __init__(self, client):
        self._client = client

    @property
    def client(self):
        """The client object bound to this cluster."""
        return self._client

    def get_info(self) -> Smap:
        """
        Returns state of AIS cluster, including the detailed information about its nodes.

        Args:
            None

        Returns:
            aistore.msg.Smap: Smap containing cluster information

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
        """
        return self.client.request_deserialize(
            HTTP_METHOD_GET,
            path="daemon",
            res_model=Smap,
            params={QParamWhat: "smap"},
        )

    def list_buckets(self, provider: str = ProviderAIS):
        """
        Returns list of buckets in AIStore cluster.

        Args:
            provider (str, optional): Name of bucket provider, one of "ais", "aws", "gcp", "az", "hdfs" or "ht".
            Defaults to "ais". Empty provider returns buckets of all providers.

        Returns:
            List[Bck]: A list of buckets

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
        """
        params = {QParamProvider: provider}
        action = ActionMsg(action=ACT_LIST).dict()

        return self.client.request_deserialize(
            HTTP_METHOD_GET,
            path="buckets",
            res_model=List[Bck],
            json=action,
            params=params,
        )

    def is_aistore_running(self) -> bool:
        """
        Returns True if cluster is ready, or false if cluster is still setting up.

        Args:
            None

        Returns:
            bool: True if cluster is ready, or false if cluster is still setting up
        """

        # compare with AIS Go API (api/cluster.go) for additional supported options
        params = {QparamPrimaryReadyReb: "true"}
        try:
            resp = self.client.request(HTTP_METHOD_GET, path="health", params=params)
            return resp.ok
        except Exception:
            return False
